"""A list of databricks reserved keywords.

https://docs.databricks.com/sql/language-manual/sql-ref-reserved-words.html
"""

RESERVED_KEYWORDS = [
    "ANTI",
    "CROSS",
    "EXCEPT",
    "FULL",
    "INNER",
    "INTERSECT",
    "JOIN",
    "LATERAL",
    "LEFT",
    "MINUS",
    "NATURAL",
    "ON",
    "RIGHT",
    "SEMI",
    "UNION",
    "USING",
]

UNRESERVED_KEYWORDS = [
    "CATALOG",
    "OPTIMIZE",
    "ZORDER",
    "PYTHON",
    "SQL",
]
